//	This code copyright (c) by the Grail project.
//	No commercial use permitted without written consent. 
//	January 1996

#ifndef		RE_CLASS_DEFINITION

enum token_type
{
	eos_t,
	error_t, 
	eset_t, 
	estring_t, 
	symbol_t, 
	plus_t, 
	star_t, 
	cat_t, 
	lparen_t, 
	rparen_t 
};

template <class S> class subexp;
template <class S> class re;
template <class S> class empty_set;
template <class S> class empty_string;
template <class S> class symbol_exp;
template <class S> class cat_exp;
template <class S> class plus_exp;
template <class S> class star_exp;
template <class S> class fm;
template <class S> class fl;

template <class S>
class subexp
{
	friend	class 	ostream;

	private:

	subexp<S>&		operator=(const subexp<S>&) { return *this; }

	public:

			subexp() { }
			subexp(const subexp<S>&) { }
	virtual 	~subexp() { }
	virtual	int	operator==(const subexp<S>*) const = 0;
	virtual	int	operator!=(const subexp<S>* s) const 
			{ return this == s ? 0 : 1;}
	virtual	int	operator<(const subexp<S>*) const = 0;
	virtual	int	operator>(const subexp<S>*) const = 0;
	virtual subexp<S>* clone() const = 0;
	virtual	int	compare_empty_string() const
			{ return 1; }
	virtual	int	compare_empty_set() const
			{ return 1; }
	virtual	int	compare_cat_exp(const subexp<S>*, const subexp<S>*) 
					const
			{ return 1; }
	virtual	int	compare_plus_exp(const subexp<S>*, const subexp<S>*) 
					const
			{ return 1; }
	virtual	int	compare_star_exp(const subexp<S>*) const
			{ return 1; }
	virtual	int	compare_symbol_exp(const S&) const
			{ return 1; }
	virtual int	contains_empty_set() const
			{ return 0; }
	virtual int	contains_empty_string() const
			{ return 0; }
	virtual void	subexp_to_fm(fm<S>&) const = 0;
	virtual void	subexp_to_fl(fl<S>&) const = 0;
	virtual int	is_empty_set() const
			{ return 0; }
	virtual int	is_empty_string() const
			{ return 0; }
	virtual int	is_finite() const
			{ return 1; }
	virtual subexp<S>*	minimize();
	virtual	void	print(ostream&, int) const
			{ }
	virtual int	size() const = 0;
	virtual	subexp<S>*	star();
};

template <class S>
class empty_set : public subexp<S>
{      
	public:
			empty_set() { }
			empty_set(const empty_set<S>&) { }
			~empty_set() { }
	empty_set<S>&	operator=(const empty_set<S>&)
			{ return *this; }
	int		operator==(const subexp<S>* s) const
			{ return s->compare_empty_set() == 0 ? 1 : 0; }
	int		operator<(const subexp<S>* s) const
			{ return s->compare_empty_set() < 0 ? 1 : 0; }
	int		operator>(const subexp<S>* s) const
			{ return s->compare_empty_set() > 0 ? 1 : 0; }
	subexp<S>*	clone() const
			{ return new empty_set<S>; }
	int		compare_empty_set() const
			{ return 0; }
	int		compare_symbol_exp(const S&) const
			{ return -1; }
	int		compare_empty_string() const
			{ return -1; }
	int		compare_cat_exp(const subexp<S>*,const subexp<S>*) const
			{ return -1; }
	int		compare_plus_exp(const subexp<S>*,const subexp<S>*) 
						const
			{ return -1; }
	int		compare_star_exp(const subexp<S>*) const
			{ return -1; }
	int		contains_empty_set() const
			{ return 1; }
	void		subexp_to_fm(fm<S>&) const; 
	int		is_empty_set() const
			{ return 1; }
#ifdef	ATT
	void		print(ostream& os,int) const;
#endif
#ifndef	ATT
	void		print(ostream& os,int) const
			{ os << re<S>::re_eset;	}
#endif
	int		size() const
			{ return 1; }
	void		subexp_to_fl(fl<S>&) const;
};

template <class S>
class empty_string : public subexp<S>
{
	public:
			empty_string() { ; }
			empty_string(const empty_string<S>&) { }
			~empty_string() { }
	empty_string<S>&	operator=(const empty_string<S>&)
			{ return *this; }
	int		operator==(const subexp<S>* s) const
			{ return s->compare_empty_string() == 0 ? 1 : 0; }
	int		operator<(const subexp<S>* s) const
			{ return s->compare_empty_string() < 0 ? 1 : 0; }
	int		operator>(const subexp<S>* s) const
			{ return s->compare_empty_string() > 0 ? 1 : 0; }
	subexp<S>*	clone() const
			{ return new empty_string<S>; }
	int		compare_empty_string() const
			{ return 0; }
	int		compare_cat_exp(const subexp<S>*,const subexp<S>*) const
			{ return -1; }
	int		compare_plus_exp(const subexp<S>*,const subexp<S>*) 
						const
			{ return -1; }
	int		compare_star_exp(const subexp<S>*) const
			{ return -1; }
	int		compare_symbol_exp(const S&) const
			{ return -1; }
	void		subexp_to_fm(fm<S>&) const; 
	int		is_empty_string() const
			{ return 1; }
#ifdef	ATT
	void		print(ostream& os,int) const;
#endif

#ifndef	ATT
	void		print(ostream& os,int) const
			{ os << re<S>::re_estring; }
#endif
	int		size() const
			{ return 1; }
	void		subexp_to_fl(fl<S>&) const;
};


template <class S>
class symbol_exp : public subexp<S>
{
#ifndef	ATT
	private:
		static pool<symbol_exp<S> >	symbol_pool;
#endif

	protected:

		S	content;	

	public:
			symbol_exp() { ; }
			symbol_exp(const S& s) : content(s) { }
			symbol_exp(const symbol_exp<S>& s) :
				content(s.content) { }
			// ~symbol_exp() {  }
	symbol_exp<S>&	operator=(const symbol_exp<S>&);
	int		operator==(const subexp<S>* s) const
			{ return s->compare_symbol_exp(content) == 0 ? 1 : 0; }
	int		operator<(const subexp<S>* s) const
			{ return s->compare_symbol_exp(content) < 0 ? 1 : 0; }
	int		operator>(const subexp<S>* s) const
			{ return s->compare_symbol_exp(content) > 0 ? 1 : 0; }
	int		compare_cat_exp(const subexp<S>*,const subexp<S>*) const
			{ return -1; }
	int		compare_plus_exp(const subexp<S>*,const subexp<S>*) 
						const
			{ return -1; }
	int		compare_star_exp(const subexp<S>*) const
			{ return -1; }
	int		compare_symbol_exp(const S& s) const;
	subexp<S>*	clone() const
			{ symbol_exp<S>* ptr = new symbol_exp<S>(content);
			  return ptr;
			}
	void		subexp_to_fm(fm<S>&) const; 
#ifndef	ATT
	void		print(ostream& os, int) const
			{ os << content; }
#endif
#ifdef	ATT
	void		print(ostream& os, int) const;
#endif
	int		size() const
			{ return 1; }
        void*           operator new(size_t);
        void            operator delete(void* p);
	void		subexp_to_fl(fl<S>&) const;
};

template <class S>
class cat_exp : public subexp<S>
{
	friend class  subexp<S>;
	friend class  re<S>;

#ifndef	ATT
	private:
		static pool<cat_exp<S> >	cat_pool;
#endif
	protected:

		subexp<S>*	left;
		subexp<S>*	right;

	public:
			cat_exp()  { left = right = 0; }
			cat_exp(const re<S>& l, const re<S>& r)
			{
				left = l.p->clone();
				right = r.p->clone();
			}
			cat_exp(const cat_exp<S>& r)
			{
				left = r.left->clone();
				right = r.right->clone();
			}
			~cat_exp() 
			{ delete left; delete right; }
	cat_exp<S>&	operator=(const cat_exp<S>&);
	int		operator==(const subexp<S>* s) const
			{ return s->compare_cat_exp(left, right) == 0 ? 1 : 0; }
	int		operator<(const subexp<S>* s) const
			{ return s->compare_cat_exp(left, right) < 0 ? 1 : 0; }
	int		operator>(const subexp<S>* s) const
			{ return s->compare_cat_exp(left, right) > 0 ? 1 : 0; }
	int		compare_cat_exp(const subexp<S>*,const subexp<S>*) 
						const;
	int		compare_plus_exp(const subexp<S>*,const subexp<S>*) 
						const
			{ return -1; }
	int		compare_star_exp(const subexp<S>*) const
			{ return -1; }
	subexp<S>*	clone() const
			{ cat_exp<S>* ptr = new cat_exp<S>;
			  ptr->left = left->clone(); 
			  ptr->right = right->clone();
			  return ptr;
			}
	int		contains_empty_set() const
			{ return left->contains_empty_set() 
				|| right->contains_empty_set(); }
	int		contains_empty_string() const
			{ return left->contains_empty_string() 
				|| right->contains_empty_string(); }
	void		subexp_to_fm(fm<S>&) const; 
	int		is_empty_set() const
			{ return left->is_empty_set() 
				|| right->is_empty_set() ? 1 : 0; }
	int		is_empty_string() const
			{ return left->is_empty_string() 
				&& right->is_empty_string() ? 1 : 0; }
	int		is_finite() const
			{ return left->is_finite() && right->is_finite() 
				? 1 : 0; }
	subexp<S>*	minimize();
	void		print(ostream&,int) const;
	int		size() const
			{ return 1 + left->size() + right->size(); }
        void*           operator new(size_t);
        void            operator delete(void* p);
	void		subexp_to_fl(fl<S>&) const;

};

template <class S>
class plus_exp : public subexp<S>
{
	friend class subexp<S>;
	friend class  re<S>;

#ifndef	ATT
	private:
		static pool<plus_exp<S> >	plus_pool;
#endif
	protected:
	
		subexp<S>*	left;
		subexp<S>*	right;

	public:
			plus_exp() { left = right = 0; }
			plus_exp(const re<S>& l, const re<S>& r)
			{
				left = l.p->clone();
				right = r.p->clone();
			}
			plus_exp(const plus_exp<S>& s)
			{
				left = s.left->clone();
				right = s.right->clone();
			}
			~plus_exp() 
			{ delete left; delete right; }
	plus_exp<S>&	operator=(const plus_exp<S>&);
	int		operator==(const subexp<S>* s) const
			{ return s->compare_plus_exp(left,right) == 0 ? 1 : 0; }
	int		operator<(const subexp<S>* s) const
			{ return s->compare_plus_exp(left,right) < 0 ? 1 : 0; }
	int		operator>(const subexp<S>* s) const
			{ return s->compare_plus_exp(left,right) > 0 ? 1 : 0; }
	int		compare_plus_exp(const subexp<S>*,const subexp<S>*) 
						const;
	int		compare_star_exp(const subexp<S>*) const
			{ return -1; }
	subexp<S>*	clone() const
			{ plus_exp<S>* ptr = new plus_exp<S>;
			  ptr->left = left->clone();
			  ptr->right = right->clone();
			  return ptr;
			}
	int		contains_empty_set() const
			{ return left->contains_empty_set() 
				|| right->contains_empty_set(); }
	int		contains_empty_string() const
			{ return left->contains_empty_string() 
				|| right->contains_empty_string(); }
	void		subexp_to_fm(fm<S>&) const;
	int		is_empty_set() const
			{ return left->is_empty_set() && 
				right->is_empty_set() ? 1 : 0; }
	int		is_empty_string() const
			{ if (left->is_empty_string())
			  { if (right->is_empty_string() || 
			        	right->is_empty_set())
				return 1; else return 0; }
			  if (left->is_empty_set() && right->is_empty_string())
				return 1;
			  return 0;
			}
	int		is_finite() const
			{ return left->is_finite() && right->is_finite() 
				? 1 : 0; }
	subexp<S>*	minimize();
	void		print(ostream&,int) const;
	int		size() const
			{ return 1 + left->size() + right->size(); }
        void*           operator new(size_t);
        void            operator delete(void* p);
	void		subexp_to_fl(fl<S>&) const;
};

template <class S>
class star_exp : public subexp<S>
{
	friend class subexp<S>;
	friend class  re<S>;

#ifndef	ATT
	private:
		static pool<star_exp<S> >	star_pool;
#endif
	protected:
	
		subexp<S>*	left;

	public:

 			star_exp() { left = 0;}
			star_exp(const re<S>& r) 
			{ left = r.p->clone(); }
			star_exp(const star_exp<S>& s) 
			{ left = s.left->clone(); }
 			~star_exp() 
			{ delete left; }
	star_exp<S>&	operator=(const star_exp<S>&);
	int		operator==(const subexp<S>* s) const
			{ if (s == 0) return 0; else return s->compare_star_exp(left) == 0 ? 1 : 0; }
	int		operator<(const subexp<S>* s) const
			{ return s->compare_star_exp(left) < 0 ? 1 : 0; }
	int		operator>(const subexp<S>* s) const
			{ return s->compare_star_exp(left) > 0 ? 1 : 0; }
	int		compare_star_exp(const subexp<S>*) const;
	subexp<S>*	clone() const
			{ star_exp<S>* ptr = new star_exp<S>;
			  ptr->left = left->clone();
			  return ptr; }
	int		contains_empty_string() const
			{ return left->contains_empty_string(); }
	int		contains_empty_set() const
			{ return left->contains_empty_set(); }
	void		subexp_to_fm(fm<S>&) const;
	int		is_empty_set() const
			{ return 0; }
	int		is_empty_string() const
			{ return left->is_empty_string() || 
				left->is_empty_set() ? 1 : 0; }
	int		is_finite() const
			{ return left->is_empty_string() || 
				left->is_empty_set() ? 1 : 0; }
#ifdef	ATT
	void		print(ostream& os,int) const;
#endif
#ifndef	ATT
	void		print(ostream& os,int) const
			{ left->print(os, 3); os << re<S>::re_star; }
#endif
	subexp<S>*	star() 
			{ return this; }
	int		size() const
			{ return 1 + left->size(); }
        void*           operator new(size_t);
        void            operator delete(void* p);
	void		subexp_to_fl(fl<S>&) const;
};

template <class S>
class re
{
	friend class istream;

	public:

		static	char	re_star;
		static	char	re_plus;
		static	char	re_cat;
		static	char	re_lparen;
		static	char	re_rparen;
		static	char*	re_estring;
		static	char*	re_eset;
		static	char	re_left_delimiter;
		static	char	re_right_delimiter;
		static	char	re_left_symbol_delimiter;
		static	char	re_right_symbol_delimiter;

	friend
	ostream&	operator<<(ostream&, const re<S>&);

	friend
	istream&	operator>>(istream&, re<S>&);

	friend		class cat_exp<S>;
	friend		class plus_exp<S>;
	friend		class star_exp<S>;

	protected:

		subexp<S>*	p;

	private:

	int		fill_buffer(istream&,char**,char,char);

	public:
			re() { p = 0; }
			re(const re<S>& r) { if (r.is_clear()) p = 0; 
				else p = r.clone();  }
			~re() 
			{ delete p; }
	re<S>&		operator=(const re<S>&); 
	re<S>&		operator=(const subexp<S>&); 
	int		operator==(const re<S>& r) const
			{ if (p == 0 && r.p == 0) return 1;
				else if (p == 0 || r.p == 0) return 0;
				else return *p == r.p ? 1 : 0; }
	int		operator!=(const re<S>& r) const
			{ if (p == 0 && r.p == 0) return 0;
				else if (p == 0 || r.p == 0) return 1;
				else return *p == r.p ? 0 : 1; }
	int		operator<(const re<S>& r) const
			{ if (p == 0 && r.p == 0) return 0;
				else if (p == 0) return 1;
				else if (r.p == 0) return 0;
				else return *p < r.p ? 1 : 0; }
	int		operator>(const re<S>& r) const
			{ if (p == 0 && r.p == 0) return 0;
				else if (p == 0) return 0;
				else if (r.p == 0) return 1;
				else return *p > r.p ? 1 : 0; }
	re<S>&		operator^=(const re<S>&);
	re<S>&		operator^=(re<S>*);
	re<S>&		operator+=(const re<S>&);
	re<S>&		operator+=(re<S>*);
	re<S>&		cat(subexp<S>*);
	re<S>&		plus(subexp<S>*);
	void		clear()
			{ delete p; p = 0; }
	subexp<S>*	clone() const
			{ return p == 0 ? 0 : p->clone(); }
	int		contains_empty_set() const
			{ return p == 0 ? 0 : p->contains_empty_set(); }
	int		contains_empty_string() const
			{ return p == 0 ? 0 : p->contains_empty_string(); }
	void		fltore(fl<S>&);
	void		fmtore(fm<S>&);
	int		is_clear() const
			{ return p == 0 ? 1 : 0; }
	int		is_empty_set() const
			{ return p == 0 ? 0 : p->is_empty_set(); }
	int		is_empty_string() const
			{ return p == 0 ? 0 : p->is_empty_string(); }
	int		is_finite() const
			{ return p == 0 ? 0 : p->is_finite(); }
	re<S>&		make_empty_set()
			{ delete p; p = new empty_set<S>; return *this; }
	re<S>&		make_empty_string()
			{ delete p; p = new empty_string<S>; return *this; }
	re<S>&		make_symbol(const S& s)
			{ delete p; p = new symbol_exp<S>(s); return *this; }
	re<S>&		minimize();
	re<S>*		parse(istream&);
	void		print(ostream& os, int) const
			{ if (p != 0) p->print(os, 0); }
	void		retofm(fm<S>&) const;
	void		retofl(fl<S>&) const;
	int		size() const 
			{ return p == 0 ? 0 : p->size(); }
	re<S>&		star();
	subexp<S>*	term(istream&);
	token_type	token(istream&);
};

#define		RE_CLASS_DEFINITION
#endif
